
#include "ppc.h"
#include "m33.h"
#include "framebuffer.h"

PPC::PPC(float hfov, int _w, int _h) : w(_w), h(_h) {

  C = V3(0.0f, 0.0f, 0.0f);
  a = V3(1.0f, 0.0f, 0.0f);
  b = V3(0.0f, -1.0f, 0.0f);
  float alpha = hfov / 2.0f * 3.1415926f / 180.0f;
  c = V3(-(float)w / 2.0f, (float)h / 2.0f, -(float)w / (2.0f *tanf(alpha)));

}


bool PPC::Project(V3 P, V3 &PP) {

  M33 cam;
  cam.SetColumn(0, a);
  cam.SetColumn(1, b);
  cam.SetColumn(2, c);

  V3 q = cam.Inverse() * (P-C);

  if (q[2] <= 0.0f)
    return false;

  PP[0] = q[0] / q[2];
  PP[1] = q[1] / q[2];
  PP[2] = 1.0f / q[2];
  return true;

}

void PPC::PositionAndOrient(V3 newC, V3 lap, V3 vpv) {


  V3 newVD = (lap - newC).UnitVector();

  V3 newa = (newVD ^ vpv).UnitVector();
  newa = newa * a.Length();

  V3 newb = (newVD ^ newa).UnitVector();
  newb = newb * b.Length();

  V3 newc = newVD*GetF() -newa*(float)w/2.0f
    -newb*(float)h/2.0f;

  a = newa;
  b = newb;
  c = newc;
  C = newC;

}

V3 PPC::GetVD() {

  V3 ret;
  ret = (a ^ b).UnitVector();
  return ret;

}

float PPC::GetF() {

  float ret;
  ret = GetVD()*c;
  return ret;

}

V3 PPC::GetPointOnImagePlane(float uf, float vf) {

  V3 ret;
  ret = C + c + a*uf + b*vf;
  return ret;

}

V3 PPC::GetPointOnFocalPlane(float uf, float vf, float f) {

  float of = GetF();
  V3 ret;
  ret = C + (c + a*uf + b*vf)*f / of;
  return ret;

}


void PPC::RenderWireframe(PPC *ppc, FrameBuffer *fb, float f,
    unsigned int color) {

  V3 baseCorners[4];
  baseCorners[0] = GetPointOnFocalPlane(0.0f, 0.0f, f);
  baseCorners[1] = GetPointOnFocalPlane(0.0f, (float)h, f);
  baseCorners[2] = GetPointOnFocalPlane((float)w, (float)h, f);
  baseCorners[3] = GetPointOnFocalPlane((float)w, 0.0f, f);
  for (int si = 0; si < 4; si++) {
    fb->Draw3DSegment(C, baseCorners[si], ppc, color);
    fb->Draw3DSegment(baseCorners[si], baseCorners[(si+1)%4], ppc, color);
  }

}

void PPC::SetByInterpolation(PPC* ppc0, PPC* ppc1, float frac) {

  V3 newC = ppc0->C + (ppc1->C - ppc0->C)*frac;
  V3 vd0 = ppc0->a ^ ppc0->b;
  V3 vd1 = ppc1->a ^ ppc1->b;
  V3 newvd = (vd0 + (vd1 - vd0)*frac).UnitVector();
  V3 vpv = (ppc0->b + (ppc1->b - ppc0->b)*frac).UnitVector()*-1.0f;
  PositionAndOrient(newC, newC + newvd, vpv);

}
